import torch
import torch.nn as nn
import torch.nn.functional as F


USE_CUDA = torch.cuda.is_available()
device = torch.device('cuda' if USE_CUDA else 'cpu')


class ActorCritic(nn.Module):
    def __init__(self, n_state, n_hidden, n_action):
        """
        Actor network, takes the input state and outputs the action probabilities
        It learns the optimal policy by updating the model using information provided by the critic
        Critic network evaluates how good it is to be at the input state by computing the value function.
        The value guides the actor on how it should adjust.
        :param n_state: number of state of the environment
        :param n_hidden: [list] number of hidden neurons in hidden layer
        :param n_action: number of actions
        """
        super(ActorCritic, self).__init__()
        self.n_state = n_state
        self.n_action = n_action
        self.input = nn.Linear(self.n_state, n_hidden[0])
        self.hidden = nn.ModuleList()
        for k in range(len(n_hidden) - 1):
            self.hidden.append(nn.Linear(n_hidden[k], n_hidden[k+1]))
        self.actor = nn.Linear(n_hidden[-1], self.n_action)
        self.critic = nn.Linear(n_hidden[-1], 1)

    def forward(self, state):
        output = state.clone().detach().requires_grad_(True)
        output = self.input(output)
        output = F.relu(output)
        for m in self.hidden:
            output = m(output)
            output = F.relu(output)
        action_probs = F.softmax(self.actor(output), dim=-1)
        state_values = self.critic(output)
        return action_probs, state_values


class PolicyNetwork(nn.Module):
    def __init__(self, n_state, n_hidden, n_action, lr=0.001):
        super(PolicyNetwork, self).__init__()
        self.model = ActorCritic(n_state, n_hidden, n_action)
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr)

    def predict(self, state):
        # Compute the output using the Actor Critic model
        # Return action probabilities, state_value
        return self.model(torch.tensor(state, dtype=torch.float32, device=device))

    def update(self, returns, log_probs, state_values):
        loss = 0
        for log_prob, value, Gt in zip(log_probs, state_values, returns):
            advantage = Gt - value.item()
            #advantage = (advantage - advantage.mean()) / advantage.std()
            policy_loss = (-log_prob * advantage)

            Gt = torch.unsqueeze(Gt, 0)
            value_loss = F.smooth_l1_loss(value, Gt)

            loss += policy_loss + value_loss

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        # self.scheduler.step()

    def get_action(self, state):
        action_probs, state_value = self.predict(state)
        action = torch.multinomial(action_probs, 1).item()
        log_prob = torch.log(action_probs[action])
        return action, log_prob, state_value


