import random
import torch
import torch.nn as nn
import numpy as np

from utils.utils import to_np

# Check if using cuda and define device
USE_CUDA = torch.cuda.is_available()
device = torch.device('cuda' if USE_CUDA else 'cpu')


#############################
# DQN model
#############################
class DQN(nn.Module):
    def __init__(self, n_state, n_action, n_hidden, lr):
        super(DQN, self).__init__()
        self.criterion = torch.nn.SmoothL1Loss()
        self.model = torch.nn.Sequential(
            torch.nn.Linear(n_state, n_hidden),
            torch.nn.ReLU(),
            torch.nn.Linear(n_hidden, n_hidden),
            torch.nn.ReLU(),
            torch.nn.Linear(n_hidden, n_hidden),
            torch.nn.ReLU(),
            torch.nn.Linear(n_hidden, n_action)
        )

        self.optimizer = torch.optim.Adam(self.model.parameters(), lr)
        # self.scheduler = torch.optim.lr_scheduler.StepLR(self.optimizer, step_size=100, gamma=0.9)

    def predict(self, state):
        """
        Compute the Q values of the state for all actions using the learning model
        @param state: input state
        @return: Q values of the state for all actions
        """
        with torch.no_grad():
            return self.model(torch.Tensor(state).to(device))

    def eps_greedy_policy(self, state, n_action, epsilon):
        """
        Compute epsilon greedy search

        Eps Greedy policy either:
        - takes a random action with probability epsilon
        - takes current best action with prob (1 - epsilon)
        @param state: state
        @param epsilon: epsison
        """
        p = random.random()
        if p < epsilon:
            # random action
            return random.randint(0, n_action - 1)
        else:
            q_values = self.predict(state)
            return torch.argmax(q_values).item()

    def boltzmann_policy(self, state, n_action, tau=0.5, clip=(-500., 500.)):
        """
        Boltzmann policy builds a probability law on q values and returns an
        action selected randomly according to this law
        :param tau: tau
        :param clip: clip
        :return: the selected action
        """
        q_values = to_np(self.predict(state))
        q_values = q_values.astype('float64')
        q_values[np.isnan(q_values)] = 0
        exp_values = np.exp(np.clip(q_values / tau, clip[0], clip[1]))
        probs = exp_values / np.sum(exp_values)
        action = np.random.choice(range(n_action), p=probs)
        return action

    def update(self, state, y):
        """
        Update the weights of the DQN given a training sample
        @param state: state
        @param y: target value
        """
        y_pred = self.model(torch.Tensor(state).to(device))
        y = torch.Tensor(y).to(device)
        loss = self.criterion(y_pred, y)
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        # self.scheduler.step()
        return loss

    def replay(self, replay_buffer, replay_batch_size, gamma):
        """
        Experience replay
        @param replay_buffer: a list of experience
        @param replay_batch_size: the number of samples we use to update the model each time
        @param gamma: the discount factor
        """
        if len(replay_buffer) >= replay_batch_size:
            replay_data = random.sample(replay_buffer, replay_batch_size)
            states = []
            td_targets = []
            for state, action, next_state, reward, is_done in replay_data:
                states.append(state)
                q_values = self.predict(state).tolist()
                if is_done:
                    q_values[action] = reward
                else:
                    q_values_next = self.predict(next_state)
                    q_values[action] = reward + gamma * torch.max(q_values_next).item()

                td_targets.append(q_values)

            loss = self.update(states, td_targets)
            return loss