import torch
import torch.nn as nn

USE_CUDA = torch.cuda.is_available()
device = torch.device('cuda' if USE_CUDA else 'cpu')


# Define the policy net
class PolicyNetwork(nn.Module):
    def __init__(self, n_state, n_hidden, n_action, lr):
        '''
        Initialize the policy neural network:
            Use one hidden layer
            Input: a state, followed by a hidden layer
            Output: the probability of taking possible individual actions
                    use softmax function as the activation for the output layer

        '''
        super(PolicyNetwork, self).__init__()
        self.network = nn.Sequential(
            nn.Linear(n_state, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, n_hidden),
            nn.Softmax(dim=-1)
        )

        self.optimizer = torch.optim.Adam(self.network.parameters(), lr)

    def predict(self, state):
        # Compute the action probabilities of state s using the learning rate
        action_probs = self.network(torch.tensor(state, dtype=torch.float32, device=device))
        return action_probs

    def update(self, advantages, log_probs):
        '''
        Update the network parameters,
            given all the data gathered in an episode, including the returns and
            the log probabilities of all steps, compute the policy gradients,
            then update the policy parameters accordingly via backpropagation

        '''
        # Update the weights of the policy network given the training samples
        # returns --> cumulative rewards for each step in an episode
        policy_gradient = []
        for log_prob, Gt in zip(log_probs, advantages):
            policy_gradient.append(-log_prob * Gt)

        loss = torch.cat(policy_gradient).sum().to(device)
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        return loss

    def get_action(self, state):
        probs = self.predict(state)
        # Sample the action with the highest probability
        action = torch.multinomial(probs, 1).item()
        log_prob = torch.log(probs[action])
        return action, log_prob


# Define the value net
class ValueNetwork(nn.Module):
    '''
        Use a regression neural network to approximate state-values
    '''

    def __init__(self, n_state, n_hidden, lr=0.01):
        super(ValueNetwork, self).__init__()
        self.criterion = torch.nn.SmoothL1Loss()
        self.model = torch.nn.Sequential(
            nn.Linear(n_state, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, 1)
        )
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr)

    def update(self, state, y):
        y_pred = self.model(torch.tensor(state, dtype=torch.float32, device=device))
        y = y.clone().requires_grad_(True).to(device)
        loss = self.criterion(y_pred, y)
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        return loss

    def predict(self, state):
        with torch.no_grad():
            return self.model(torch.tensor(state, dtype=torch.float32, device=device))
