import gym
from gym import spaces
import numpy as np
from gym.utils import seeding


class Wendtris(gym.Env):
    """
    Custom Environment that follows gym interface
    """
    metadata = {'render.modes': ['human']}

    def __init__(self, NUM_ORDERS=20, NUM_SPACEWIDTH=6, NUM_CAPACITY=6, PENALTY_FACTOR=5):
        super(Wendtris, self).__init__()

        self.seed()
        self.num_spacewidth = NUM_SPACEWIDTH
        self.num_capacity = NUM_CAPACITY
        self.num_orders = NUM_ORDERS
        self.num_remain_orders = NUM_ORDERS
        self.penalty_factor = PENALTY_FACTOR
        self.action_space = spaces.Discrete(2)
        self.reset()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        seed += 1
        return [seed]

    def reset(self):
        # Initialize state to the first order
        # State:
        #     0 --> capacity_list
        #     1 --> current order
        #     2 --> current reward
        #     3 --> total reward
        #     4 --> remaining orders
        self.profit = 0
        self.orders, self.rewards = self.spielGenerator(self.num_orders, self.num_spacewidth)
        self.orders = self.orders.astype('float64')
        # Off-set row for orders and rewards
        orders_offset_row = np.zeros(self.orders.shape[1], dtype=np.float64)
        rewards_offset_row = np.zeros(self.rewards.shape[1])

        self.orders = np.vstack((self.orders, orders_offset_row))
        self.rewards = np.vstack((self.rewards, rewards_offset_row))

        self.order_pos = 0
        self.reward_pos = 0

        self.count_no_capacity = 0

        self.accepted_orders = []
        self.rejected_orders = []

        self.num_remain_orders = self.num_orders

        self.capacity_list = np.ones((self.num_spacewidth,), dtype=np.float64) * self.num_capacity

        self.tmp = np.array(
            [self.capacity_list, self.orders[self.order_pos], self.rewards[self.reward_pos][0], self.num_remain_orders],
            dtype=object)

        self.state = np.hstack(self.tmp.flatten())

        return self.state

    def spielGenerator(self, NUM_ORDERS, NUM_SPACEWIDTH, REWARD_MIN_PER_UNIT=1, REWARD_MAX_PER_UNIT=9):

        base_element_set = np.array(
            # Data based on Wendtris Java code (Hagen)
            [[1, 2, 3, 2, 1, 0],
             [1, 1, 1, 1, 0, 0],
             [0, 2, 2, 0, 0, 0],
             [2, 2, 1, 1, 0, 0],
             [1, 3, 2, 2, 1, 0],
             [2, 1, 1, 1, 3, 3]], dtype=object)

        ARR_ORDERLIST = np.zeros((NUM_ORDERS, NUM_SPACEWIDTH), dtype=np.int8)

        ARR_SET_AND_POSITION = np.zeros((NUM_ORDERS, 3), dtype=np.int8)
        # 0 -> element (index of base_element_set)
        # 1 -> position (offset)
        # 2 -> length of element

        ARR_REWARDS = np.zeros((NUM_ORDERS, 2), dtype=np.float16)
        # 0 -> reward per unit
        # 1 -> total reward

        # Index 0: Set index
        ARR_SET_AND_POSITION[:, 0] = self.np_random.randint(base_element_set.shape[0], size=(NUM_ORDERS))

        # Set reward data per unit
        # ARR_REWARDS_CLASSES = np.arange(REWARD_MIN_PER_UNIT, REWARD_MAX_PER_UNIT + 1, 5)
        ARR_REWARDS[:, 0] = self.np_random.randint(REWARD_MIN_PER_UNIT, REWARD_MAX_PER_UNIT + 1, size=(NUM_ORDERS))

        # Index 1: Position of element
        for j in range(NUM_ORDERS):
            # Set radon position for the elements
            len_element = len(base_element_set[ARR_SET_AND_POSITION[j, 0]])
            pos_element = self.np_random.randint(NUM_SPACEWIDTH - len_element + 1)
            ARR_SET_AND_POSITION[j, 1] = pos_element
            ARR_SET_AND_POSITION[j, 2] = len_element

            # set reward data for total element
            ARR_REWARDS[j, 1] = ARR_REWARDS[j, 0] * sum(base_element_set[ARR_SET_AND_POSITION[j, 0]])

            for k in range(len_element):
                ARR_ORDERLIST[j, pos_element + k] = base_element_set[ARR_SET_AND_POSITION[j, 0]][k]

        return ARR_ORDERLIST, ARR_REWARDS

    def step(self, action):
        assert self.action_space.contains(action)
        # state
        # 0: capacity list
        # 1: current order
        # 2: current unit reward
        # 3: number of remaining orders
        if action:
            # If the agent accepts the order#
            no_capacity = ((self.tmp[0] - self.tmp[1]) < 0).any()
            if no_capacity:
                # If there is no capacity, agent receives negative current reward * penalty_factor
                # Order position and remaining orders will not change
                current_reward = - self.rewards[self.reward_pos][1] * self.penalty_factor
                self.profit += current_reward
                self.order_pos += 1
                self.reward_pos += 1
                self.num_remain_orders -= 1
                self.tmp[2] = self.rewards[self.reward_pos][0]
                self.tmp[3] = self.num_remain_orders
                self.count_no_capacity += 1
            else:
                current_reward = self.rewards[self.reward_pos][1]
                self.profit += current_reward
                self.tmp[0] -= self.tmp[1]
                self.accepted_orders.append(self.order_pos)
                self.order_pos += 1
                self.reward_pos += 1
                self.num_remain_orders -= 1
                self.tmp[2] = self.rewards[self.reward_pos][0]
                self.tmp[3] = self.num_remain_orders

            self.tmp[1] = self.orders[self.order_pos]

        else:
            # If the agent rejects the order
            current_reward = 0
            self.profit += current_reward
            self.num_remain_orders -= 1
            self.rejected_orders.append(self.order_pos)
            self.order_pos += 1
            self.reward_pos += 1
            self.tmp[1] = self.orders[self.order_pos]
            self.tmp[2] = self.rewards[self.reward_pos][0]
            self.tmp[3] = self.num_remain_orders

        done = bool(
            self.num_remain_orders == 0)

        info = {"Number no capacity": self.count_no_capacity,
                "Accepted orders": self.accepted_orders,
                "Rejected orders": self.rejected_orders,
                "Total rewards": self.profit}

        self.state = np.hstack(self.tmp.flatten())

        return np.hstack(self.state.flatten()), current_reward, done, info


class Wendtris_Eva(gym.Env):
    # Evaluation environment
    def __init__(self, orders, rewards, num_spacewidth=6, num_capacity=6):
        self.orders = orders.astype('float64')
        self.rewards = rewards
        self.num_spacewidth = num_spacewidth
        self.num_capacity = num_capacity
        self.action_space = spaces.Discrete(2)
        orders_offset_row = np.zeros(self.orders.shape[1], dtype=np.float64)
        rewards_offset_row = np.zeros(self.rewards.shape[1])
        self.num_remain_orders = self.orders.shape[0]

        self.orders = np.vstack((self.orders, orders_offset_row))
        self.rewards = np.vstack((self.rewards, rewards_offset_row))

        self.order_pos = 0
        self.reward_pos = 0

        self.count_no_capacity = 0

        self.accepted_orders = []
        self.rejected_orders = []

        self.capacity_list = np.ones((self.num_spacewidth,), dtype=np.float64) * self.num_capacity

        self.tmp = np.array(
            [self.capacity_list, self.orders[self.order_pos], self.rewards[self.reward_pos][0], self.num_remain_orders],
            dtype=object)

        self.state = np.hstack(self.tmp.flatten())

    def step(self, action):
        assert self.action_space.contains(action)
        # state
        # 0: capacity list
        # 1: current order
        # 2: current unit reward
        # 3: number of remaining orders
        if action:
            # If the agent accepts the order#
            no_capacity = ((self.tmp[0] - self.tmp[1]) < 0).any()
            if no_capacity:
                # If there is no capacity, agent receives 0 reward
                # Order position and remaining orders will not change
                current_reward = 0
                self.order_pos += 1
                self.reward_pos += 1
                self.num_remain_orders -= 1
                self.tmp[2] = self.rewards[self.reward_pos][0]
                self.tmp[3] = self.num_remain_orders
                self.count_no_capacity += 1
            else:
                current_reward = self.rewards[self.reward_pos][1]
                self.tmp[0] -= self.tmp[1]
                self.accepted_orders.append(self.order_pos)
                self.order_pos += 1
                self.reward_pos += 1
                self.num_remain_orders -= 1
                self.tmp[2] = self.rewards[self.reward_pos][0]
                self.tmp[3] = self.num_remain_orders

            self.tmp[1] = self.orders[self.order_pos]

        else:
            # If the agent rejects the order
            current_reward = 0
            self.num_remain_orders -= 1
            self.rejected_orders.append(self.order_pos)
            self.order_pos += 1
            self.reward_pos += 1
            self.tmp[1] = self.orders[self.order_pos]
            self.tmp[2] = self.rewards[self.reward_pos][0]
            self.tmp[3] = self.num_remain_orders

        done = bool(
            self.num_remain_orders == 0)

        info = {"Number no capacity": self.count_no_capacity,
                "Accepted orders": self.accepted_orders,
                "Rejected orders": self.rejected_orders}

        self.state = np.hstack(self.tmp.flatten())

        return np.hstack(self.state.flatten()), current_reward, done, info




