import errno
import os
import shutil
import sys
import math
import pickle
import numpy as np


def to_np(var):
    return var.detach().cpu().numpy()


def sliding_window(data, N):
    """
    For each index, k, in data we average over the window from k-N-1 to k. The beginning handles incomplete buffers,
    that is it only takes the average over what has actually been seen.
    :param data: A numpy array, length M
    :param N: The length of the sliding window.
    :return: A numpy array, length M, containing smoothed averaging.
    """
    idx = 0
    window = np.zeros(N)
    smoothed = np.zeros(len(data))

    for i in range(len(data)):
        window[idx] = data[i]
        idx += 1

        smoothed[i] = window[0:idx].mean()

        if idx == N:
            window[0:-1] = window[1:]
            idx = N - 1

    return smoothed


def create_folder(folder_path):
    """Create a folder if it does not exist.
    """
    try:
        os.makedirs(folder_path)
    except OSError as _e:
        if _e.errno != errno.EEXIST:
            raise


def clear_folder(folder_path):
    """Clear all contents recursively if the folder exists.
    Create the folder if it has been accidently deleted.
    """
    if not os.path.exists(folder_path):
        create_folder(folder_path)
    for the_file in os.listdir(folder_path):
        _file_path = os.path.join(folder_path, the_file)
        try:
            if os.path.isfile(_file_path):
                os.unlink(_file_path)
            elif os.path.isdir(_file_path):
                shutil.rmtree(_file_path)
        except OSError as _e:
            print(_e)


def save_list(data, folder_path, filename):
    """
    Save a list variable to file
    @folder_path: str
    @file_name: str
    """
    if type(filename) != str:
        raise Exception("The input is not a string")
    if not os.path.exists(folder_path):
        create_folder(folder_path)
    file_path = os.path.join(folder_path, filename + '.data')
    with open(file_path, 'wb') as filehandle:
        pickle.dump(data, filehandle)


def load_list(folder_path):
    with open(folder_path, 'rb') as filehandle:
        return pickle.load(filehandle)


def modify_orders(orders):
    """Modify the orders for evaluation"""
    # optimal orders contains the index of accepted orders, -->1 accepted
    orders = [sorted(i, reverse=False) for i in orders]     # sort the orders first
    result = []
    for indexes in orders:
        tmp = [0] * 20
        for i in indexes:
            tmp[i] = 1
        result.append(tmp)
    return np.asarray(result, dtype=int)


def stretched_exponential_decay(episode, n_episode, a=0.2, b=0.1, c=0.1):
    """
    Stretched exponential decay for epsilon greedy algorthm
    :param time: current episode
    :param n_episode: the number of episodes
    :param a:  decides where the agent spend more time <.5 more exploiting >.5 more exploration
    :param b:  decides the slope of transition region between exploration and exploitation
    :param c:  controls the steepness of left and right tail of the graph
    """
    standardized_time = (episode - a * n_episode) / (b * n_episode)
    cosh = np.cosh(math.exp(-standardized_time))
    epsilon = 1.1 - (1 / cosh + (episode * c / n_episode))
    return epsilon


def anneal_tau(time, decay_rate, tau):
    """
    Decay tau value over time
    :rtype: float
    """
    return np.exp(-decay_rate * time) * tau + 1


class StdOut(object):
    """Redirect stdout to file, and print to console as well.
    """
    def __init__(self, output_file):
        self.terminal = sys.stdout
        self.log = open(output_file, "a")

    def write(self, message):
        self.terminal.write(message)
        self.terminal.flush()
        self.log.write(message)
        self.log.flush()

    def flush(self):
        self.terminal.flush()
        self.log.flush()